/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file gabor.cc
   \brief 2D & 3D Gabor features.
	
	This file contains functions for calculating 2D/3D Gabor features.

	B. S. Manjunath, P. Salembier, and T. Sikora, editors. Introduction to
	MPEG-7: Multimedia Content Description Interface. Wiley & Sons, Inc.,
	New York, USA, Apr. 2002.
	
	T. Majtner, D. Svoboda. Texture Analysis Using 3D Gabor Features and
	3D MPEG-7 Edge Histogram Descriptor in Fluorescence Microscopy. In
	4th International Conference on 3D Imaging (IC3D). Los Alamitos, California,
	USA: IEEE Computer Society, 2014. Pages 17, ISBN 978-1-4799-8023-9.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include <fstream>
#include <string> 
#include <iostream>
#include <complex>
#include <math.h>

#include "descriptors/gabor.h"
#include "transform.h"
#include "convolution.h"

using namespace std;

namespace i3d {


template <class T> void ComputeGabor2D(Image3d<T> &img, int SCALES, int ORIENTATIONS, int KERNEL_SIZE, vector<double> &final_vector)
{	

	double fmax = 0.25;  // together with parameter alpha define frequency
	double fu, alpha, tetav, xprime, yprime;
	std::complex<double> i_unit(0, 1);

	Image3d<float> gaborFilter, filteredImage, img2;

	// preparing variables/room for the convolution
	gaborFilter.MakeRoom(KERNEL_SIZE, KERNEL_SIZE, 1);
	filteredImage.MakeRoom(img.GetSizeX(), img.GetSizeY(), 1);
	img2.MakeRoom(img.GetSizeX(), img.GetSizeY(), 1);
	
	// deriving mean value of original image as the first component of feature vector
	double mean_orig = 0.0, std_orig = 0.0;
	for (unsigned ii = 0; ii < img.GetSizeX(); ++ii)
	{
		for (unsigned jj = 0; jj < img.GetSizeY(); ++jj)
		{
			mean_orig += img.GetVoxel(ii, jj, 0);
			img2.SetVoxel(ii, jj, 0, img.GetVoxel(ii, jj, 0));
		}
	}
	mean_orig = mean_orig / (img.GetSizeX() * img.GetSizeY());

	// deriving standard deviation value of original image as the second component of feature vector
	for (unsigned ii = 0; ii < img.GetSizeX(); ++ii)
	{
		for (unsigned jj = 0; jj < img.GetSizeY(); ++jj)
		{
			std_orig += pow(img.GetVoxel(ii, jj, 0) - mean_orig, 2.0);
		}
	}
	std_orig = sqrt(std_orig / (img.GetSizeX() * img.GetSizeY()));

	// print mean and standard deviation
	final_vector.push_back(mean_orig);
	final_vector.push_back(std_orig);

	// main cycle of the algorithm
	for (int i = 1; i <= SCALES; i++)
	{
		fu = fmax / pow((sqrt(2.0)), (i-1));
		alpha = fu / sqrt(2.0);
		gaborFilter.Clear();
    
		for (int j = 1; j <= ORIENTATIONS; j++)
		{
			tetav = (double(j-1) / ORIENTATIONS) * M_PI;
        
			for (int x = 1; x <= KERNEL_SIZE; x++)
			{
				for (int y = 1; y <= KERNEL_SIZE; y++)
				{
					// deriving gabor filter
					xprime =  (x-((KERNEL_SIZE+1)/2.0)) * cos(tetav) + (y-((KERNEL_SIZE+1)/2.0)) * sin(tetav);
					yprime = -(x-((KERNEL_SIZE+1)/2.0)) * sin(tetav) + (y-((KERNEL_SIZE+1)/2.0)) * cos(tetav);
					gaborFilter.SetVoxel(x-1, y-1, 0, (float) real((pow(fu, 2.0) / (M_PI * sqrt(2.0) * sqrt(2.0))) * 
						            exp(-((pow(alpha, 2.0)) * (pow(xprime, 2.0)) + (pow(alpha, 2.0)) * (pow(yprime, 2.0)))) * 
									exp(i_unit * (double)2 * M_PI * fu * xprime)));
				}
			}

			#ifdef I3D_DEBUG
				// save Gabor filter with scale i and orientation j
				if ((i == 4) && (j == 5)) gaborFilter.SaveImage("_filter.ics");
			#endif					

			// performing convolution of input image with derived gabor filter
			filteredImage.Clear();
			Convolution<float, float>(img2, gaborFilter, filteredImage, false);

			#ifdef I3D_DEBUG
				// save filtered image
				if ((i == 4) && (j == 5)) filteredImage.SaveImage("_output.ics");
			#endif

			// deriving mean value of filtered image
			double mean = 0.0, std = 0.0;			
			for (unsigned ii = 0; ii < filteredImage.GetSizeX(); ++ii)
			{
				for (unsigned jj = 0; jj < filteredImage.GetSizeY(); ++jj)
				{
					mean += filteredImage.GetVoxel(ii, jj, 0);
				}
			}
			mean = mean / (filteredImage.GetSizeX() * filteredImage.GetSizeY());

			// deriving standard deviation of filtered image
			for (unsigned ii = 0; ii < filteredImage.GetSizeX(); ++ii)
			{
				for (unsigned jj = 0; jj < filteredImage.GetSizeY(); ++jj)
				{
					std += pow(filteredImage.GetVoxel(ii, jj, 0) - mean, 2.0);
				}
			}
			std = sqrt(std / (filteredImage.GetSizeX() * filteredImage.GetSizeY()));

			// print mean and standard deviation of filtered image
			final_vector.push_back(mean);
			final_vector.push_back(std);
		}
	}
}


template <class T> void ComputeGabor3D(Image3d<T> &img, int SCALES, int ORIENTATIONS, int KERNEL_SIZE, vector<double> &final_vector)
{	
	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();

	 if ((res.x != res.y) || (res.x != res.z))
	 {
		  float max_res = std::max(std::max(res.x, res.y), res.z);
		  Vector3d<float> new_res = max_res;
		  ResampleToDesiredResolution(img, new_res, LANCZOS);
	 }

	int KERNEL_SIZE_Z = (int) floor(KERNEL_SIZE / ((double)std::max(img.GetSizeX(), img.GetSizeY()) / img.GetSizeZ()));
	KERNEL_SIZE_Z = std::max(1, KERNEL_SIZE_Z);

	double fmax = 0.25;  // together with parameter alpha define frequency
	double fu, alpha, theta, phi, xprime, yprime, zprime;
	std::complex<double> i_unit(0, 1);

	// preparing variables/room for the convolution
	Image3d<float> gaborFilter, filteredImage, img2;	
	gaborFilter.MakeRoom(KERNEL_SIZE, KERNEL_SIZE, KERNEL_SIZE_Z);
	filteredImage.MakeRoom(img.GetSizeX(), img.GetSizeY(), img.GetSizeZ());
	img2.MakeRoom(img.GetSizeX(), img.GetSizeY(), img.GetSizeZ());	
	
	// deriving mean value of original image as the first component of feature vector
	double mean_orig = 0.0, std_orig = 0.0;
	for (unsigned ii = 0; ii < img.GetSizeX(); ++ii)
	{
		for (unsigned jj = 0; jj < img.GetSizeY(); ++jj)
		{
			for (unsigned kk = 0; kk < img.GetSizeZ(); ++kk)
			{
				mean_orig += img.GetVoxel(ii, jj, kk);
				img2.SetVoxel(ii, jj, kk, img.GetVoxel(ii, jj, kk));
			}
		}
	}
	mean_orig = mean_orig / img.GetImageSize();

	// main cycle of the algorithm
	for (int i = 1; i <= SCALES; i++)
	{
		fu = fmax / pow((sqrt(2.0)), (i-1));
		alpha = fu / sqrt(2.0);

		for (int k = 1; k <= ORIENTATIONS; k++)
		{		
			phi = (double(k-1) / ORIENTATIONS) * M_PI;

			for (int j = 1; j <= ORIENTATIONS; j++)
			{
				theta = (double(j-1) / ORIENTATIONS) * M_PI / 2;
				gaborFilter.Clear();
        
				for (int x = 1; x <= KERNEL_SIZE; x++)
				{
					for (int y = 1; y <= KERNEL_SIZE; y++)
					{
						for (int z = 1; z <= KERNEL_SIZE_Z; z++)
						{
							// deriving 3D gabor filter
							yprime = (y-((KERNEL_SIZE+1)/2.0)) * cos(theta) - 
									 (z-((KERNEL_SIZE_Z+1)/2.0)) * sin(theta);
							zprime = (y-((KERNEL_SIZE+1)/2.0)) * sin(theta) + 
									 (z-((KERNEL_SIZE_Z+1)/2.0)) * cos(theta);
							xprime = (x-((KERNEL_SIZE+1)/2.0)) * cos(phi) - yprime * sin(phi);
							yprime = (x-((KERNEL_SIZE+1)/2.0)) * sin(phi) + yprime * cos(phi);
							gaborFilter.SetVoxel(x-1, y-1, z-1, (float) real((pow(fu, 2.0) / (M_PI * 2)) * 
										exp(-((pow(alpha, 2.0)) * (pow(xprime, 2.0)) + 
											  (pow(alpha, 2.0)) * (pow(yprime, 2.0)) + 
											  (pow(alpha, 2.0)) * (pow(zprime, 2.0)))) * 
										exp(i_unit * (double)2 * M_PI * fu * xprime)));
						}
					}
				}

				#ifdef I3D_DEBUG
					// save Gabor filter with scale i and orientation j, k
					if ((i == 1) && (j == 1) && (k == 1)) gaborFilter.SaveImage("_filter.ics");
				#endif	

				// performing convolution of input image with derived gabor filter
				filteredImage.Clear();
				Convolution<float>(img2, gaborFilter, filteredImage, false);

				#ifdef I3D_DEBUG
					// save filtered image
					if ((i == 1) && (j == 1) && (k == 1)) filteredImage.SaveImage("_output.ics");
				#endif	

				// deriving mean value of filtered image
				double mean = 0.0, std = 0.0;
				for (unsigned ii = 0; ii < filteredImage.GetSizeX(); ++ii)
				{
					for (unsigned jj = 0; jj < filteredImage.GetSizeY(); ++jj)
					{
						for (unsigned kk = 0; kk < filteredImage.GetSizeZ(); ++kk)
						{
							mean += filteredImage.GetVoxel(ii, jj, kk);
						}
					}
				}
				mean = mean / img.GetImageSize();

				// deriving standard deviation of filtered and original image
				for (unsigned ii = 0; ii < filteredImage.GetSizeX(); ++ii)
				{
					for (unsigned jj = 0; jj < filteredImage.GetSizeY(); ++jj)
					{
						for (unsigned kk = 0; kk < filteredImage.GetSizeZ(); ++kk)
						{
							std += pow(filteredImage.GetVoxel(ii, jj, kk) - mean, 2.0);
							if ((i == 1) && (j == 1) && (k == 1)) 
								std_orig += pow(img.GetVoxel(ii, jj, kk) - mean_orig, 2.0);
						}
					}
				}				

				if ((i == 1) && (j == 1) && (k == 1)) 
				{
					std_orig = sqrt(std_orig / img.GetImageSize());

					// print mean and standard deviation of original image
					final_vector.push_back(mean_orig);
					final_vector.push_back(std_orig);
				}

				std = sqrt(std / img.GetImageSize());

				// print mean and standard deviation of filtered image
				final_vector.push_back(mean);
				final_vector.push_back(std);

				// for elimination angle combinations projected to xy plane
				if (k==1) break;
			}
		}
	}
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template I3D_ALGO_EXPORT void ComputeGabor2D<GRAY8>(Image3d<GRAY8> &, int, int, int, vector<double> &);
template I3D_ALGO_EXPORT void ComputeGabor2D<GRAY16>(Image3d<GRAY16> &, int, int, int, vector<double> &);

template I3D_ALGO_EXPORT void ComputeGabor3D<GRAY8>(Image3d<GRAY8> &, int, int, int, vector<double> &);
template I3D_ALGO_EXPORT void ComputeGabor3D<GRAY16>(Image3d<GRAY16> &, int, int, int, vector<double> &);

}
