/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file lbpplaton.cc
   \brief 3D LBP-platonic texture features.
	
	This file contains functions for calculating 3D LBP-platonic features.

	[1] T. Majtner, D. Svoboda. Comparison of 3D Texture-based Image Descriptors
	in Fluorescence Microscopy. In 16th InternationalWorkshop on Combinatorial
	Image Analysis (IWCIA). Lecture Notes in Computer Science, Volume
	8466. Switzerland: Springer International Publishing, 2014. Pages 186195,
	ISBN 978-3-319-07147-3.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include <fstream>
#include <math.h>

#include "descriptors/lbpplaton.h"
#include "transform.h"

// default value from [1]
#define R 1			// radius

using namespace std;

namespace i3d {

template <class T> void ComputeLBP_PLATON(const Image3d<T> &img_orig, int SOLID, vector<double> &feature_vector)
{
	Image3d<T> img = img_orig;

	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();
	if ((res.x != res.y) || (res.x != res.z))
	{
		float max_res = std::max(std::max(res.x, res.y), res.z);
		Vector3d<float> new_res = max_res;
		ResampleToDesiredResolution(img, new_res, LANCZOS);
	}	

	int hist_len = 0;  // length of the histogram

	// for setting the right length of the histogram based on the number of vertices
	switch (SOLID) {
	  case 1:
		  hist_len = (int)pow(2.0, 4);   // tetrahedron
		  break;
	  case 2:
		  hist_len = (int)pow(2.0, 8);   // hexahedron
		  break;
	  case 3:
		  hist_len = (int)pow(2.0, 6);   // octahedron
		  break;
	  case 4:
		  hist_len = (int)pow(2.0, 12);  // icosahedron
		  break;
	  default:
		  cerr << "Unknown platonic solid, ending the program.";
		  exit(2);
	}

	int *result = new int[hist_len];

	// value is used to form LBP pattern, 
	// a1 and a2 are temp variables for defining position of examined point in space
	int value, a1, a2;
	double interpolated_value;

	for (int i = 0; i < hist_len; i++) result[i] = 0;

	for (unsigned z = 2; z < img.GetSizeZ() - 2; z++)
	{
		for (unsigned y = 2; y < img.GetSizeY() - 2; y++)
		{			
			for (unsigned x = 2; x < img.GetSizeX() - 2; x++)
			{	
				if (img.GetVoxel(x, y, z) > 0)
				{
					value = 0;
					switch (SOLID) {
					  
					case 1:
						// 4 points (values) for tetrahedron
						a1 = (int)floor(R/sqrt(2.0)); a2 = (int)ceil(R/sqrt(2.0));
						interpolated_value = img.GetVoxel(R + x, y,-a1 + z) + (img.GetVoxel(R + x, y,-a2 + z) - 
							img.GetVoxel(R + x, y,-a1 + z)) * ((R/sqrt(2.0)) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 0;
						interpolated_value = img.GetVoxel(-R + x, y,-a1 + z) + (img.GetVoxel(-R + x, y,-a2 + z) - 
							img.GetVoxel(-R + x, y,-a1 + z)) * ((R/sqrt(2.0)) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 1;

						interpolated_value = img.GetVoxel(x, R + y, a1 + z) + (img.GetVoxel(x, R + y, a2 + z) - 
							img.GetVoxel(x, R + y, a1 + z)) * ((R/sqrt(2.0)) - a1) / (a2 - a1);							
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 2;
						interpolated_value = img.GetVoxel(x,-R + y, a1 + z) + (img.GetVoxel(x,-R + y, a2 + z) - 
							img.GetVoxel(x,-R + y, a1 + z)) * ((R/sqrt(2.0)) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 3;
						break;
					  
					case 2:
						// 8 points (values) for hexahedron
						value += ((img.GetVoxel(-R + x,-R + y,-R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 0;
						value += ((img.GetVoxel(-R + x,-R + y, R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 1;
						value += ((img.GetVoxel(-R + x, R + y,-R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 2;
						value += ((img.GetVoxel(-R + x, R + y, R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 3;

						value += ((img.GetVoxel(R + x,-R + y,-R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 4;
						value += ((img.GetVoxel(R + x,-R + y, R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 5;
						value += ((img.GetVoxel(R + x, R + y,-R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 6;
						value += ((img.GetVoxel(R + x, R + y, R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 7;
						break;
					  
					case 3:
						// 6 points (values) for octahedron
						value += ((img.GetVoxel(-R + x, y, z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 0;
						value += ((img.GetVoxel( R + x, y, z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 1;
						value += ((img.GetVoxel( x,-R + y, z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 2;
						value += ((img.GetVoxel( x, R + y, z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 3;
						value += ((img.GetVoxel( x, y,-R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 4;
						value += ((img.GetVoxel( x, y, R + z) < img.GetVoxel(x, y, z)) ? 0 : 1) << 5;
						break;
					  
					case 4:
						// 12 points (values) for icosahedron
						a1 = (int)floor((((1.0 + sqrt(5.0)) / 2.0) * R)); 
						a2 = (int)ceil((((1.0 + sqrt(5.0)) / 2.0) * R));
						interpolated_value = img.GetVoxel(x,-R + y,-a1 + z) + (img.GetVoxel(x,-R + y,-a2 + z) - 
							img.GetVoxel(x,-R + y,-a1 + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 0;
						interpolated_value = img.GetVoxel(x, R + y,-a1 + z) + (img.GetVoxel(x, R + y,-a2 + z) - 
							img.GetVoxel(x, R + y,-a1 + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 1;

						interpolated_value = img.GetVoxel(x,-R + y, a1 + z) + (img.GetVoxel(x,-R + y, a2 + z) - 
							img.GetVoxel(x,-R + y, a1 + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 2;						
						interpolated_value = img.GetVoxel(x, R + y, a1 + z) + (img.GetVoxel(x, R + y, a2 + z) - 
							img.GetVoxel(x, R + y, a1 + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 3;

						interpolated_value = img.GetVoxel(-R + x,-a1 + y, z) + (img.GetVoxel(-R + x,-a2 + y, z) - 
							img.GetVoxel(-R + x,-a1 + y, z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 4;
						interpolated_value = img.GetVoxel( R + x,-a1 + y, z) + (img.GetVoxel( R + x,-a2 + y, z) - 
							img.GetVoxel( R + x,-a1 + y, z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 5;

						interpolated_value = img.GetVoxel(-R + x, a1 + y, z) + (img.GetVoxel(-R + x, a2 + y, z) - 
							img.GetVoxel(-R + x, a1 + y, z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 6;						
						interpolated_value = img.GetVoxel( R + x, a1 + y, z) + (img.GetVoxel( R + x, a2 + y, z) - 
							img.GetVoxel( R + x, a1 + y, z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 7;

						interpolated_value = img.GetVoxel(-a1 + x, y,-R + z) + (img.GetVoxel(-a2 + x, y,-R + z) - 
							img.GetVoxel(-a1 + x, y,-R + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 8;
						interpolated_value = img.GetVoxel(a1 + x, y,-R + z) + (img.GetVoxel(a2 + x, y,-R + z) - 
							img.GetVoxel(a1 + x, y,-R + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 9;

						interpolated_value = img.GetVoxel(-a1 + x, y, R + z) + (img.GetVoxel(-a2 + x, y, R + z) - 
							img.GetVoxel(-a1 + x, y, R + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 10;						
						interpolated_value = img.GetVoxel(a1 + x, y, R + z) + (img.GetVoxel(a2 + x, y, R + z) - 
							img.GetVoxel(a1 + x, y, R + z)) * ((((1.0 + sqrt(5.0)) / 2.0) * R) - a1) / (a2 - a1);
						value += ((interpolated_value < img.GetVoxel(x, y, z)) ? 0 : 1) << 11;
						break;
					  
					default:
						cerr << "Unknown platonic solid, ending the program.";
						exit(3);
					}
					result[value]++;
				}
			}
		}
	}

	int sum = 0;
	for (int i = 0; i < hist_len; i++) sum += result[i];  // for normalization
	for (int i = 0; i < hist_len; i++) feature_vector.push_back((double)(result[i]) / sum);
	
	delete[] result;
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template I3D_ALGO_EXPORT void ComputeLBP_PLATON<GRAY8>(const Image3d<GRAY8> &, int, vector<double> &);
template I3D_ALGO_EXPORT void ComputeLBP_PLATON<GRAY16>(const Image3d<GRAY16> &, int, vector<double> &);

}
