/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file vlbp.cc
   \brief 3D Volume LBP features.

	This file contains functions for calculating 3D Volume LBP features.

	[1] G. Zhao and M. Pietikinen. Dynamic Texture Recognition Using Volume
	Local Binary Patterns. In ECCV 2006 Workshop on Dynamical Vision,
	pages 165177. Springer, 2006.

	[2] M. Pietikinen, A. Hadid, G. Zhao, and T. Ahonen. Computer Vision Using
	Local Binary Patterns. Computational imaging and vision. Springer
	Verlag, London, 2011.	
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include <fstream>
#include <math.h>
#include "transform.h"
#include "descriptors/vlbp.h"



// default values from [1]
#define P 4			// number of sampling points around the circle
#define R 1			// radius

using namespace std;

namespace i3d {

template <class T> void ComputeVLBP(const Image3d<T> &img_orig, vector<double> &final_vector)
{  
	Image3d<T> img = img_orig;
	
	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();
	if ((res.x != res.y) || (res.x != res.z))
	{
		float max_res = std::max(std::max(res.x, res.y), res.z);
		Vector3d<float> new_res = max_res;
		ResampleToDesiredResolution(img, new_res, LANCZOS);
	}

	int hist_len = (int) pow(2.0, 14);  // length of the histogram
	int *result = new int[hist_len];    // for storing LBP pattern
	int pattern, bit_indicator;

	// clear the histogram
	for (int i = 0; i < hist_len; i++) result[i] = 0;

	for (unsigned z = R; z < img.GetSizeZ() - R; z++)
	{
		for (unsigned y = R; y < img.GetSizeY() - R; y++)
		{
			for (unsigned x = R; x < img.GetSizeX() - R; x++)
			{
				if (img.GetVoxel(x,y,z) != 0)
				{
					pattern = 0; bit_indicator = 1;

					pattern += ((img.GetVoxel(x, y, z-1) < img.GetVoxel(x, y, z)) ? 0 : 1) << 0;

					// using three 2D slices for derive 3D pattern  
					for (int zz = -1; zz < 2; zz++)
					{
						for (int p = 0; p < P; p++)
						{
							// finding x and y coordinates of point in the neighbourhood
							double xp = x + (((R*cos(2*M_PI*p/P) > -0.05) && (R*cos(2*M_PI*p/P) < 0.05)) ? 0 : R*cos(2*M_PI*p/P));
							double yp = y - (((R*sin(2*M_PI*p/P) > -0.05) && (R*sin(2*M_PI*p/P) < 0.05)) ? 0 : R*sin(2*M_PI*p/P));
		
							// if we are right in the centre of the pixel, we can take its value
							if ((xp == (int)xp) || (yp == (int)yp))
							{
								pattern += ((img.GetVoxel((int)(xp + 0.1), (int)(yp + 0.1), z+zz) < 
									img.GetVoxel(x, y, z)) ? 0 : 1) << bit_indicator;							
							}
							else  // if we are not, we need to use bilinear interpolation to get the value
							{
								int x1 = (int)floor(xp), x2 = (int)ceil(xp), y1 = (int)floor(yp), y2 = (int)ceil(yp);
								pattern += (((img.GetVoxel(x1,y1,z+zz)*(x2-xp)*(y2-yp) +
											img.GetVoxel(x2,y1,z+zz)*(xp-x1)*(y2-yp) +
											img.GetVoxel(x1,y2,z+zz)*(x2-xp)*(yp-y1) +
											img.GetVoxel(x2,y2,z+zz)*(xp-x1)*(yp-y1)) < 
											img.GetVoxel(x, y, z)) ? 0 : 1) << bit_indicator;
							}
							bit_indicator++;
						}
					}
					pattern += ((img.GetVoxel(x, y, z+1) < img.GetVoxel(x, y, z)) ? 0 : 1) << bit_indicator;

					result[pattern]++;
				}
			}
		}	
	}

	// for normalization
	int sum = 0;
	for (int i = 0; i < hist_len; i++) sum += result[i];
	for (int i = 0; i < hist_len; i++) final_vector.push_back((double)(result[i]) / sum);
	
	delete[] result;	
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template I3D_ALGO_EXPORT void ComputeVLBP<GRAY8>(const Image3d<GRAY8> &, vector<double> &);
template I3D_ALGO_EXPORT void ComputeVLBP<GRAY16>(const Image3d<GRAY16> &, vector<double> &);

}
