/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * heap.cc
 *
 * Min-heap data structure
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "basic.h"
#include "heap.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                     Class MinHeap                           //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		template <class KEY, class NODE> const Type::Size MinHeap<KEY, NODE>::NOT_ON_HEAP = -1;		

		/**********************************************************************/

		template <class KEY, class NODE> NODE *MinHeap<KEY, NODE>::Pop()
		{
			Type::Size num = m_nodes.size();

			if (num == 0)
			{
				throw i3d::InternalException("Cannot pop from the empty heap!");
			}
			else
			{
				NODE *root = m_nodes[0];

				if (num == 1)
				{
					// the heap contains only one node
					m_back_pointers[root -> m_index] = NOT_ON_HEAP;
					m_nodes.pop_back();
				}
				else
				{
					// move the last node to the root
					m_back_pointers[root -> m_index] = NOT_ON_HEAP;
					m_back_pointers[m_nodes[num - 1] -> m_index] = 0;
					m_nodes[0] = m_nodes[num - 1];
					m_nodes.pop_back();
					BalanceDown(0);
				}

				return root;
			}
		}

		/**********************************************************************/

		template <class KEY, class NODE> void MinHeap<KEY, NODE>::Insert(NODE *node)
		{
			Type::Size pos = m_nodes.size();
			m_nodes.push_back(node);
			m_back_pointers[node -> m_index] = pos;
			BalanceUp(pos);
		}

		/**********************************************************************/

		template <class KEY, class NODE> void MinHeap<KEY, NODE>::Update(Type::Size i, KEY value)
		{
			if (m_nodes[i] -> Update(value))
			{
				if (i == 0 || Compare((i - 1) / 2, i))
				{
					BalanceDown(i);
				}
				else
				{
					BalanceUp(i);
				}
			}
		}

		/**********************************************************************/

		template <class KEY, class NODE> void MinHeap<KEY, NODE>::BalanceUp(Type::Size i)
		{
			Type::Size parent = (i - 1) / 2;

			while (i > 0 && Compare(i, parent))
			{
				// swap the i-th node with its parent
				NODE *pom = m_nodes[i];
				m_nodes[i] = m_nodes[parent];
				m_nodes[parent] = pom;

				// swap their back-pointers
				m_back_pointers[m_nodes[i] -> m_index] = i;
				m_back_pointers[m_nodes[parent] -> m_index] = parent;

				// update indices
				i = parent;
				parent = (i - 1) / 2;
			}
		}

		/**********************************************************************/

		template <class KEY, class NODE> void MinHeap<KEY, NODE>::BalanceDown(Type::Size i)
		{
			Type::Size num = m_nodes.size();   
			Type::Size sec = 2 * i + 2; // index of the second child of the node

			while (sec < num)
			{
				if (Compare(sec - 1, sec)) // select the smaller child
				{
					--sec;
				}

				if (Compare(sec, i))
				{
					// swap the node with the smaller child
					NODE *pom = m_nodes[sec];
					m_nodes[sec] = m_nodes[i];
					m_nodes[i] = pom;

					// swap their back-pointers
					m_back_pointers[m_nodes[sec] -> m_index] = sec;
					m_back_pointers[m_nodes[i] -> m_index] = i;

					// update indices
					i = sec;
					sec = 2 * i + 2;
				}
				else
				{
					// the node is smaller than its both childs
					return;
				}
			}

			if (sec == num)  
			{
				// the node has only left child
				--sec;

				if (Compare(sec, i))
				{
					// swap the node with the left child
					NODE *pom = m_nodes[sec];
					m_nodes[sec] = m_nodes[i];
					m_nodes[i] = pom;

					// swap their back-pointers
					m_back_pointers[m_nodes[sec] -> m_index] = sec;
					m_back_pointers[m_nodes[i] -> m_index] = i;
				}
			}
		}

		/**********************************************************************/

		template <class KEY, class NODE> void MinHeap<KEY, NODE>::Clear()
		{
			// clear the list of nodes
			Type::Size num = m_nodes.size();

			for (Type::Size i = 0; i < num; ++i)
			{
				delete m_nodes[i];
			}
		}

		/**********************************************************************/

		template class MinHeap<Type::Prec, NHA_Node>;

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

