/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nha_graph.cc
 *
 * Graph data structure intended for fast detection of simple points in 3D
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "nha_graph.h"

namespace i3d
{
	namespace fls 
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                      Class NHA_Graph                        //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/
		template <int M, int Degree> bool NHA_Graph<M, Degree>::AccessibleNeighbours(int start, int num)
		{
			// begin breadth-first search algorithm from the starting vertex
			m_visited[start] = true;
			m_accessible[0] = start;

			// the number of visited vertices in a given graph
			int num_visited = 1;

			// the index to the list of accessible vertices
			int index = 0;

			// execute breadth-first search algorithm
			while (index < num_visited && num_visited < num)
			{
				// the first unprocessed accessible vertex
				const Vertex &vertex = m_vertices[m_accessible[index]];

				// go through its adjacent vertices
				for (int i = 0; i < vertex.degree; ++i)
				{
					// the index of the i-th adjacent vertex
					start = vertex.adj[i];

					// check whether the i-th adjacent vertex has already been visited
					if (!m_visited[start])
					{
						// add the i-th adjacent vertex to the list of accessible vertices
						m_visited[start] = true;
						m_accessible[num_visited] = start;
						++num_visited;
					}
				}

				// move to the next accessible vertex
				++index;
			}

			return (num_visited == num);
		}

		/**********************************************************************/

		template <int M, int Degree> bool NHA_Graph<M, Degree>::AccessibleSixNeighbours(int start, int num)
		{
			// begin breadth-first search algorithm from the starting vertex
			m_visited[start] = true;
			m_accessible[0] = start;

			// the number of visited vertices
			int num_visited = 1;

			// the index to the list of accessible vertices
			int index = 0;

			// decrement the number of 6-neighbouring vertices accessible from the starting one
			--num;

			// execute breadth-first search algorithm
			while (index < num_visited && num > 0)
			{
				// the first unprocessed accessible vertex
				const Vertex &vertex = m_vertices[m_accessible[index]];

				// go through its adjacent vertices
				for (int i = 0; i < vertex.degree; ++i)
				{
					// the index of the i-th adjacent vertex
					start = vertex.adj[i];

					// check whether the i-th adjacent vertex has already been visited
					if (!m_visited[start])
					{
						// add the i-th adjacent vertex to the list of accessible vertices
						m_visited[start] = true;
						m_accessible[num_visited] = start;
						++num_visited;

						// check whether the i-th adjacent vertex is a 6-neighbour
						if (start == 2 || start == 6 || start == 8 || start == 9 || start == 11 || start == 15)
						{
							--num;
						}
					}
				}

				// move to the next accessible vertex
				++index;
			}

			return (num == 0);
		}

		/**********************************************************************/

		template class NHA_Graph<18, 4>;
		template class NHA_Graph<26, 16>;

		/**********************************************************************/	

	} // namespace fls 

} // namespace i3d
