/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nha_routine.cc
 *
 * Routine for the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "nha_routine.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                     Class NHA_Routine                       //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		NHA_Routine::~NHA_Routine()
		{
			delete m_initializer;
			delete m_neighbourhood;
			delete m_speed_function;
			delete m_curvature;
		}

		/**********************************************************************/

		void NHA_Routine::Initialize()
		{
			std::vector<Type::Size> initial_interface;
			m_initializer -> Initialize(initial_interface);

			for (Type::Size i = 0; i < initial_interface.size(); ++i)
			{
				m_curvature -> Estimate(initial_interface[i]);
				m_base -> m_heap.Insert(new NHA_Node(initial_interface[i], 0, m_speed_function -> Get(initial_interface[i])));
			}
		}

		/**********************************************************************/

		void NHA_Routine::PerformIteration()
		{
			NHA_Node *min = m_base -> m_heap.Pop();

			if (min -> m_speed > 0)
			{
				Type::Size num_neigh = 0;
				const Type::Size *neigh = m_neighbourhood -> ExternalNeighbours(min -> m_index, num_neigh);

				m_base -> m_implicit_function[min -> m_index] = -1;
				m_curvature -> Increment(min -> m_index);

				for (Type::Size i = 0; i < num_neigh; ++i)
				{
					m_base -> m_implicit_function[neigh[i]] = 0;
					m_curvature -> Estimate(neigh[i]);
					m_base -> m_heap.Insert(new NHA_Node(neigh[i], min -> m_key, m_speed_function -> Get(neigh[i])));
				}
			}
			else
			{
				Type::Size num_neigh = 0;
				const Type::Size *neigh = m_neighbourhood -> InternalNeighbours(min -> m_index, num_neigh);

				m_base -> m_implicit_function[min -> m_index] = 1;

				for (Type::Size i = 0; i < num_neigh; ++i)
				{
					m_base -> m_implicit_function[neigh[i]] = 0;
					m_curvature -> Estimate(neigh[i]);
					m_curvature -> Decrement(neigh[i]);
					m_base -> m_heap.Insert(new NHA_Node(neigh[i], min -> m_key, m_speed_function -> Get(neigh[i])));
				}
			}

			delete min;
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

