

#include "metrics.h"

namespace i3d{
	using namespace std;

	Image3d<double> ssim(Image3d<double> &imageOriginal, Image3d<double> &imageDistorted, bool bit16,
						double gaussSigma, float gaussSize, double constant1, double constant2)
	{
		//setting max intensity for constant computation
		double maxIntens = 255;
		if(bit16){
			maxIntens = 65535;
		}

		//computation of constants
		double c1 = pow(constant1*maxIntens,2.0);
		double c2 = pow(constant2*maxIntens,2.0);
		
		//variable initiation
		Image3d<double> muOi = imageOriginal;//original image intensity gaussed and squared (later has right value)
		Image3d<double> muDi = imageDistorted;//distorted image intensity gaussed and squared (later has right value)

		Image3d<double> muOD = imageOriginal;//multiplied orig. and dist. image intensity average
		Image3d<double> oiSquared = imageOriginal;//image intensities squared
		Image3d<double> diSquared = imageDistorted;
		Image3d<double> odSquared = imageOriginal;

		//computation of Gauss
		GaussFIR(muOi,gaussSigma,gaussSize);
		GaussFIR(muDi,gaussSigma,gaussSize);

		//setting right value to variables
		for(unsigned int z=0; z<imageOriginal.GetSizeZ();z++){
			for(unsigned int y=0; y<imageOriginal.GetSizeY();y++){
				for(unsigned int x=0; x<imageOriginal.GetSizeX();x++){
					double muOiVoxel = muOi.GetVoxel(x,y,z);
					double muDiVoxel = muDi.GetVoxel(x,y,z);
					double oiVoxel = oiSquared.GetVoxel(x,y,z);
					double diVoxel = diSquared.GetVoxel(x,y,z);
					
					muOD.SetVoxel(x,y,z,
						muOiVoxel*muDiVoxel);
					muOi.SetVoxel(x,y,z,
						pow(muOiVoxel, 2.0));
					muDi.SetVoxel(x,y,z,
						pow(muDiVoxel, 2.0));
					oiSquared.SetVoxel(x,y,z,
						pow(oiVoxel, 2.0));
					diSquared.SetVoxel(x,y,z,
						pow(diVoxel, 2.0));
					odSquared.SetVoxel(x,y,z,
						oiVoxel*diVoxel);
				}
			}
		}//oiSquared, diSquared, odSquared and muOD are set to the right value.
		 //muOi and muDi are squared.
		
		//setting *Squared to contrast value of images
		GaussFIR(oiSquared,gaussSigma,gaussSize);
		GaussFIR(diSquared,gaussSigma,gaussSize);
		GaussFIR(odSquared,gaussSigma,gaussSize);//squared are now sigma (in MSSIM equation)

		//initiation of result variable. ImageOriginal is set so that result has right dimensions.
		Image3d<double> result = imageOriginal;
		
		//final computation
		for(unsigned int z=0; z<imageOriginal.GetSizeZ();z++){
			for(unsigned int y=0; y<imageOriginal.GetSizeY();y++){
				for(unsigned int x=0; x<imageOriginal.GetSizeX();x++){
					double muOiVoxel = muOi.GetVoxel(x,y,z);
					double muDiVoxel = muDi.GetVoxel(x,y,z);
					double muODVoxel = muOD.GetVoxel(x,y,z);//mu in MSSIM equation
					
					double sigmaOiVoxel = oiSquared.GetVoxel(x,y,z)-muOiVoxel;
					double sigmaDiVoxel = diSquared.GetVoxel(x,y,z)-muDiVoxel;
					double sigmaODVoxel = odSquared.GetVoxel(x,y,z)-muODVoxel;//sigma in MSSIM equation


					double numerator = 0;
					double denumerator = 0;

					numerator = ((2*muODVoxel)+c1)*((2*sigmaODVoxel)+c2);
					denumerator = (muOiVoxel+muDiVoxel+c1)*(sigmaOiVoxel+sigmaDiVoxel+c2);//MSSIM equation

					result.SetVoxel(x,y,z,
								numerator/denumerator);//setting ssim map
				}
			}
		}

		return result;
	}

	template<class VOXEL>
	double Mssim (const Image3d<VOXEL> imageOriginal, const Image3d<VOXEL> imageDistorted,
					double gaussSigma, float gaussSize, double constant1, double constant2)
	{
		//image dimension comparison
		if(imageOriginal.GetSize() != imageDistorted.GetSize())
		{
			throw IOException("Images don't have same resolution");
		}
		//variable initiation
		Image3d<double> originalImage;
		Image3d<double> distortedImage;
		
		Image3d<double> ssimMap;
		double result = 0;

		//conversion to double precision
		originalImage.ConvertFrom<double,VOXEL>(imageOriginal);
		distortedImage.ConvertFrom<double,VOXEL>(imageDistorted);
		
		//creation of ssim map and different constants if image is in 16 or 8 bits
		if(typeid(imageOriginal)==typeid(Image3d<GRAY8>)){
			ssimMap = ssim(originalImage,distortedImage, false, gaussSigma, gaussSize, constant1, constant2);
		}else{
			ssimMap = ssim(originalImage,distortedImage, true, gaussSigma, gaussSize, constant1, constant2);
		}

		//computation of mean of ssim map
		for(unsigned int i=0;i<ssimMap.GetSizeZ();i++){
			double result_slice=0;
			for(unsigned int j=0;j<ssimMap.GetSizeY();j++){
				double result_row = 0;
				for(unsigned int k=0;k<ssimMap.GetSizeX();k++){
					result_row += ssimMap.GetVoxel(k,j,i);
				}
				result_slice += result_row;
			}
			result += result_slice;
		}
		result /=ssimMap.GetImageSize();

		return result;

	}
	
	template<class VOXEL>
	double MSE (const Image3d<VOXEL> imageOriginal, const Image3d<VOXEL> imageDistorted, bool linearStrech)
	{
		//image dimension comparison
		if(imageOriginal.GetSize() != imageDistorted.GetSize())
		{
			throw IOException("Images don't have same resolution");
		}

		//variable initiation
		valarray<double> origData;
		valarray<double> disoData;

		Image3d<double> orig;
		Image3d<double> dist;

		//conversion to double precision
		orig.ConvertFrom<double,VOXEL>(imageOriginal);
		dist.ConvertFrom<double,VOXEL>(imageDistorted);
		origData = orig.GetVoxelData();
		disoData = dist.GetVoxelData();


		//linear strech
		if(linearStrech){
			double scale = (origData.max()-origData.min())/(disoData.max()-disoData.min());
			disoData -= disoData.min();
			disoData *= scale;
			disoData += origData.min();
		}

		//MSE computation
		valarray<double> result = origData-disoData;

		result *= result;

		return result.sum()/result.size();
	}

	template<class VOXEL>
	double PSNR (const Image3d<VOXEL> imageOriginal, const Image3d<VOXEL> imageDisorted, bool linearStrech)
	{
		double result = 0;
		double maxImageIntensity = imageOriginal.GetMaxValue();

		result = log10( pow(maxImageIntensity,2.0)
						/ MSE(imageOriginal, imageDisorted, linearStrech));

		return 10*result;
	}

	template<class VOXEL>
	double FSC (const Image3d<VOXEL> imageOriginal, const Image3d<VOXEL> imageDistorted)
	{
		//comparison of image dimensions
		if(imageOriginal.GetSize() != imageDistorted.GetSize())
		{
			throw IOException("Images don't have same resolution");
		}

		//variable initiation
		complex<double> numerator(0,0);
		double denumeratorOI=0;
		double denumeratorDI=0;

		Image3d< complex<double> > fourierOrig;
		Image3d< complex<double> > fourierDist;

		//Fourier transform
		FFT(imageOriginal, fourierOrig, true);
		FFT(imageDistorted, fourierDist, true);
		
		//execution of fsc equation which is available on wikipedia
		for(unsigned int z=0;z<imageOriginal.GetSizeZ();z++){
			for(unsigned int y=0; y<imageOriginal.GetSizeY();y++){
				for(unsigned int x=0;  x<(imageOriginal.GetSizeX()/2)+1;x++){//FFT in i3d creates only half size complex image
					if(x==0){
						numerator += fourierOrig.GetVoxel(x,y,z)*conj(fourierDist.GetVoxel(x,y,z));
						denumeratorOI += pow(abs(fourierOrig.GetVoxel(x,y,z)),2.0);
						denumeratorDI += pow(abs(fourierDist.GetVoxel(x,y,z)),2.0);
					}else{
						numerator += fourierOrig.GetVoxel(x,y,z)*conj(fourierDist.GetVoxel(x,y,z));
						denumeratorOI += pow(abs(fourierOrig.GetVoxel(x,y,z)),2.0);
						denumeratorDI += pow(abs(fourierDist.GetVoxel(x,y,z)),2.0);

						numerator += conj(fourierOrig.GetVoxel(x,y,z))*fourierDist.GetVoxel(x,y,z);
						denumeratorOI += pow(abs(conj(fourierOrig.GetVoxel(x,y,z))),2.0);
						denumeratorDI += pow(abs(conj(fourierDist.GetVoxel(x,y,z))),2.0);
					}
				}
			}
		}
		
		return abs(numerator/sqrt(denumeratorDI*denumeratorOI));
	}

	//--------------------------------------------------------------------------
	// explicit instantiation
	//--------------------------------------------------------------------------

	template double Mssim (const Image3d<GRAY8>, const Image3d<GRAY8>, double, float, double, double);
	template double Mssim (const Image3d<GRAY16>, const Image3d<GRAY16>, double, float, double, double);
	template double Mssim (const Image3d<float>, const Image3d<float>, double, float, double, double);
	template double Mssim (const Image3d<double>, const Image3d<double>, double, float, double, double);
	template double Mssim (const Image3d<int>, const Image3d<int>, double, float, double, double);

	template double MSE (const Image3d<GRAY8>, const Image3d<GRAY8>, bool);
	template double MSE (const Image3d<GRAY16>, const Image3d<GRAY16>, bool);
	template double MSE (const Image3d<float>, const Image3d<float>, bool);
	template double MSE (const Image3d<double>, const Image3d<double>, bool);
	template double MSE (const Image3d<int>, const Image3d<int>, bool);

	template double PSNR (const Image3d<GRAY8>, const Image3d<GRAY8>, bool);
	template double PSNR (const Image3d<GRAY16>, const Image3d<GRAY16>, bool);
	template double PSNR (const Image3d<float>, const Image3d<float>, bool);
	template double PSNR (const Image3d<double>, const Image3d<double>, bool);
	template double PSNR (const Image3d<int>, const Image3d<int>, bool);

	template double FSC (const Image3d<GRAY8>, const Image3d<GRAY8>);
	template double FSC(const Image3d<GRAY16>, const Image3d<GRAY16>);
	template double FSC(const Image3d<float>, const Image3d<float>);
	template double FSC(const Image3d<double>, const Image3d<double>);
}
