/*
 * segment_nuclei.cpp
 *
 * Function for segmentation of nuclei
 *
 * Marek Vinkler (xvinkl@mail.muni.cz) 2009
 * 
 */

#include "segment_nuclei.h"
#include "regions.h"
#include "threshold.h"
#include "LevelSet.h"
#include "morphology.h"
#include "bin.h"
#include "rank_filters.h"

namespace i3d
{
	bool SegmentNuclei2d(const Image3d<i3d::GRAY16> &in, Image3d<i3d::GRAY16> &out, const SNR_Params &params)
	{
		Image3d<i3d::GRAY16> mask_img, label_gray, c1, c2;
		Image3d<bool> lap_bin, gm_bin, mask_bin, bin, bin_small, merged_mask;
		Image3d<float> gauss, lap_img, gm_img;
		LabeledImage3d<i3d::GRAY16, bool> label;
		Histogram hist;
	
		GrayToFloat(in, gauss);
		GaussIIR(gauss, (float)params.laplace_sigma);
		Laplace(gauss, lap_img); // Other i3d Laplacians are unsigned
		Threshold(lap_img, lap_bin, -std::numeric_limits<float>::infinity(), 0.0f);

		GrayToFloat(in, gauss);
		GaussIIR(gauss, (float)params.laplace_sigma);
		EstimateGradient2_2D(gauss, gm_img, false);

		switch(params.grad_thresh_type)
		{
		case Background:
			IntensityHist(gm_img, hist);
			Threshold(gm_img, gm_bin, (float) FindUnimodalThreshold(hist));
			break;

		case Median:
			gm_bin.CopyMetaData(gm_img);
			Threshold(gm_img, gm_bin, gm_img.GetKPercentileValue(50.0f));
			break;

		case Border: break;
		
		// TODO later
		//case Border:
		//	/*border = dilation(lap_img,2)-erosion(lap_img, 2);
		//	border_lbl = label(border,2);
		//	border_msr = measure(border_lbl, gradmag(img, params.gradmag_sigma), 'Mean');
		//	gm_img = msr2obj(border_lbl, border_msr);
		//	gm_img = gm_img > params.min_avg_gradmag;*/
		//	Image3d<bool> dilation, erosion, border;

		//	Dilation(lap_bin, dilation, 2);
		//	Erosion(lap_bin, erosion, 2);
		//	for(size_t i = 0; i < dilation.GetSizeX()*dilation.GetSizeY(); i++)
		//	{
		//		border.SetVoxel(i, dilation.GetVoxel(i) > erosion.GetVoxel(i));
		//	}
		//	label.CreateRegions(border, nb2D_o8);
	        //	break;
		}

		if(params.remove_faint_objects)
		{
			IntensityHist(in, hist);
			Threshold(in, mask_bin, (GRAY16) FindUnimodalThreshold(hist));
			bool *lapB = lap_bin.GetFirstVoxelAddr();
			bool *gmB = gm_bin.GetFirstVoxelAddr();
			bool *maskB = mask_bin.GetFirstVoxelAddr();
			for(size_t i = 0; i < lap_bin.GetSizeX()*lap_bin.GetSizeY(); i++)
			{
				*maskB = *lapB & *gmB & *maskB;
				lapB++;
				gmB++;
				maskB++;
			}
		}
		else
		{
			mask_bin.CopyMetaData(lap_bin);
			bool *lapB = lap_bin.GetFirstVoxelAddr();
			bool *gmB = gm_bin.GetFirstVoxelAddr();
			bool *maskB = mask_bin.GetFirstVoxelAddr();
			for(size_t i = 0; i < lap_bin.GetSizeX()*lap_bin.GetSizeY(); i++)
			{
				*maskB = *lapB & *gmB;
				lapB++;
				gmB++;
				maskB++;
			}
		}

		label.CreateForegroundRegionsFF(mask_bin, nb2D_o8);

		// Delete components with volume < params.min_fragment_size
		Remove(label.components, std::binder2nd<component_volume_less<GRAY16, bool> >
			(component_volume_less<GRAY16, bool>(), params.min_fragment_size));

		label.Convert(label_gray);
		Threshold(label_gray, mask_bin, (GRAY16) 1);

		// The neighbourhood variant doesn't work
		Reconstruction_by_dilation(mask_bin, lap_bin, bin);
		
		label.CreateForegroundRegionsFF(bin, nb2D_o8);

		// Delete components with volume < params.min_fragment_size
		Remove(label.components, std::binder2nd<component_volume_less<GRAY16, bool> >
			(component_volume_less<GRAY16, bool>(), params.min_fragment_size));

		// COMMENT: Ignore params.closing_size for now
		label.Convert(label_gray);
		Closing(label_gray, c1, nb2D_o8);
		//label.components.begin()->second.voxel_value = label.get_max_label()+1;
		label.components.begin()->second.voxel_value = true;
		label.Convert(label_gray);
		Opening(label_gray, c2, nb2D_o8);

		i3d::GRAY16 *labelG = label_gray.GetFirstVoxelAddr();
		i3d::GRAY16 *c1G = c1.GetFirstVoxelAddr();
		i3d::GRAY16 *c2G = c2.GetFirstVoxelAddr();
		for(size_t i = 0; i < label_gray.GetSizeX()*label_gray.GetSizeY(); i++)
		{
			if(*c1G == *c2G)
				*labelG = *c1G;
			else
				*labelG = 0;

			labelG++;
			c1G++;
			c2G++;
		}

		Threshold(label_gray, bin, (GRAY16) 1);
		label.CreateForegroundRegionsFF(bin, nb2D_o8);
		Remove(label.components, component_touch_border<GRAY16, bool>());
		label.Convert(label_gray);
		Threshold(label_gray, bin, (GRAY16) 1);
		BinFillHoles(bin, nb2D_o8);

		if(params.merge_small_objects)
		{
			label.CreateForegroundRegionsFF(bin, nb2D_o8);
			// Delete components with volume > params.max_fragment_size
			Remove(label.components, std::binder2nd<component_volume_greater<GRAY16, bool> >
				(component_volume_greater<GRAY16, bool>(), params.max_fragment_size));
			label.Convert(label_gray);
			Threshold(label_gray, bin_small, (GRAY16) 1);
			Closing(bin_small, merged_mask, (float) params.closing_size);

			bool *binB = bin.GetFirstVoxelAddr();
			bool *mergedB = merged_mask.GetFirstVoxelAddr();
			for(size_t i = 0; i < bin.GetSizeX()*bin.GetSizeY(); i++)
			{
				*binB = *binB | *mergedB;
				binB++;
				mergedB++;
			}
		}

		label.CreateForegroundRegionsFF(bin, nb2D_o8);
		// Delete components with volume < params.min_cell_size
		Remove(label.components, std::binder2nd<component_volume_less<GRAY16, bool> >
			(component_volume_less<GRAY16, bool>(), params.min_cell_size));

		label.Convert(out);

		return true;
	}
}  // namespace i3d

