/*
* i3dlib - image manipulation library
*
* Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, write to the Free
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/** \file watershed_contour_dynamics.cc
\brief Part of mathematical morphology routines - implementation.

This file contains bodies of mathematical morphology routines dealing with 
the dynamics of the watershed contours principle by Najman and Schmitt.
*/ 

#include <set>
#include <map>
#include <vector>
#include <limits>
#ifdef I3D_DEBUG
#include <iostream>
#endif // #ifdef I3D_DEBUG

#include "watersheds.h"
#include "i3dassert.h"


#ifdef __GNUG__
#	pragma implementation
#endif

namespace i3d {

	//--------------------------------------------------------------------------
	// Watershed contour dynamics implementation
	//--------------------------------------------------------------------------

	/// handles all the operations necessary for the watershed contour dynamics computation (is able to create the flooding list and to derive contour dynamics from it)
	template<typename T>
	class CWSCDBasinManager
	{
	private:
		/// holds all the information about one basin
		struct SBasin
		{
			/// minimum of the basin
			T minimum;
			/// dynamics of this basin computed when it has been flooded by some other basin
			T dynamics;
			/// pointer to the basin with the lower minimum that flooded this one
			SBasin *floodedBy;
			/// helper flag used while searching for the common flooder of some two basins
			bool visited;

			/// initializes new basin on specified level
			SBasin(): minimum(std::numeric_limits<T>::max()), dynamics(std::numeric_limits<T>::max()), floodedBy(NULL), visited(false) {}

			/// returns by witch basin is this one flooded (recursively - the one with the highest dynamics)
			SBasin* FloodedBy()
			{
				if (floodedBy)
					return floodedBy->FloodedBy();
				else
					return this;
			}

			/// returns by witch basin is this one flooded (recursively) requesting minimum dynamics value
			SBasin* FloodedBy(T minDyn)
			{
				// allow the same dynamics just in the case we whoud have to return noone - be careful while connecting the tree not to connect some node to itself
				if (minDyn < dynamics)
					return this;
				else
				{
					if (!floodedBy)
					{
						if (minDyn == dynamics)
							return this;
						else
							return NULL;
					}
					else
						return floodedBy->FloodedBy(minDyn);
				}
			}

			/// returns predecessor of given basin, it also clears the visited flag
			SBasin* GetPredecessorOf(SBasin *basin)
			{
				visited = false;
				// when we are the predecessor, return the result
				if (floodedBy == basin)
					return this;
				// stop walking when we are looking for the predecessor of ourselves (resulting dynamics should be from the other path in the flooding list) 
				if (this == basin)
					return NULL;
				// ok, let's look deeper
				if (floodedBy)
					return floodedBy->GetPredecessorOf(basin);
				I3DASSERT_BT(false, "Someone wanted the predecessof of something that was not on this path");
				return NULL;
			}

			/// clears the visited flag as long as it is set while walking the flooding list
			void ClearVisited()
			{
				visited = false;
				if (floodedBy && floodedBy->visited)
					floodedBy->ClearVisited();				
			}
		};

		/// map of the labels to the basins
		std::map<size_t, SBasin*> basins;
		/// map of the label pairs to the saddle points and later to the dynamics (transformation in the ComputeFloodingList method - saddle points are forgotten and default dynamics are set)
		std::map<std::pair<size_t, size_t>, T> basinPairsSPDyn;
		// TODO try hash map

		/// returns basin of given label
		SBasin* GetBasin(size_t label)
		{
			SBasin *result = basins[label];
			if (!result)
			{
				result = new SBasin();
				basins[label] = result;
			}
			return result;
		}

		/// computes basin dynamics for one of the basins when two basins meet and creates the part of the flooding list (should be called from the lowest saddle points to the highest ones)
		void ComputeDynamics(size_t label1, size_t label2, T level)
		{
			I3DASSERT_BT(label1 != label2, "Trying to compute the dynamics for the two same basins");

			SBasin *p0 = GetBasin(label1);
			SBasin *q0 = GetBasin(label2);

			// find members of connected componentes with the highest dynamics and do the computations for them
			SBasin *pn, *qn, *flooding0, *floodedn;
			pn = p0->FloodedBy();
			qn = q0->FloodedBy();
			// when two parts of the one connected component meet again, we do not need to do anything
			if (pn == qn)
				return;
			// decide who floods who - the basin with the lower minimum floods the one with the higher minimum
			if (pn->minimum < qn->minimum)
			{
				floodedn = qn;
				flooding0 = p0;
			}
			else if (pn->minimum > qn->minimum)
			{
				floodedn = pn;
				flooding0 = q0;
			}
			else
			{
				// let's do the choice that when two basins have the same minimum, the one with the lower label floods the one with the higher label
				if (label1 < label2)
				{
					floodedn = qn;
					flooding0 = p0;
				}
				else
				{
					floodedn = pn;
					flooding0 = q0;
				}
			}
			// compute the new dynamics value for the flooded bysin and interconnect the flooding list
			I3DASSERT_BT(floodedn->dynamics == std::numeric_limits<T>::max(), "Flooded basin has already some dynamics value - this should never happen");
			I3DASSERT_BT(floodedn->floodedBy == NULL, "Flooded basin has already been flooded by someone before - this should never happen");
			T floodedDynamics = level - floodedn->minimum; 
			SBasin *qj = flooding0->FloodedBy(floodedDynamics); // the basin which flooded the floodedn (basin from the other connected component with the dynamics higher than the new computed one but as low as possible)
			I3DASSERT_BT(qj, "Something is wrong with the so far created flooding list");
			I3DASSERT_BT(qj != floodedn, "Something is wrong with the so far created flooding list (it is not a tree?)");
			floodedn->dynamics = floodedDynamics;
			floodedn->floodedBy = qj;
		}

	public:
		/// frees all the allocated resources
		~CWSCDBasinManager()
		{
			for (typename std::map<size_t, SBasin*>::iterator i = basins.begin(); i != basins.end(); ++i)
				delete i->second;
		}

		/// corrects minima value for some basin if necessary
		void PotentialMinimum(size_t label, T value)
		{
			SBasin *basin = GetBasin(label);
			if (basin->minimum > value)
				basin->minimum = value;
		}

		/// corrects saddle point value for some basins if necessary
		void PotentialSaddlePoint(T value, const std::set<size_t> &labels)
		{
			for (std::set<size_t>::const_iterator first = labels.begin(); first != --labels.end(); ++first)
			{
				std::set<size_t>::const_iterator second = first;
				for (++second; second != labels.end(); ++second)
				{
					// try to lookup the saddle point
					typename std::map<std::pair<size_t, size_t>, T>::iterator saddlePoint = basinPairsSPDyn.find(std::pair<size_t, size_t>(*first, *second));
					// if no saddle point candidate ready, insert new one
					if (saddlePoint == basinPairsSPDyn.end())
						basinPairsSPDyn.insert(std::pair<std::pair<size_t, size_t>, T>(std::pair<size_t, size_t>(*first, *second), value));
					// correct the saddle point candidate if necessary
					else
						if (value < saddlePoint->second)
							saddlePoint->second = value;
				}
			};
		}

		/// computes the flooding list from the know saddle points (information about the saddle points is forgotten and default (uninitialized) dynamics value is stored)
		void ComputeFloodingList()
		{
			// sort the saddle points from the lowest ones
			std::vector<std::pair<T, std::pair<size_t, size_t> > > sortedSaddlePoints;
			sortedSaddlePoints.reserve(basinPairsSPDyn.size());
			for (typename std::map<std::pair<size_t, size_t>, T>::iterator i = basinPairsSPDyn.begin(); i != basinPairsSPDyn.end(); ++i)
			{
				sortedSaddlePoints.push_back(std::pair<T, std::pair<size_t, size_t> >(i->second, i->first));
				i->second = T(0);
			}
			std::sort(sortedSaddlePoints.begin(), sortedSaddlePoints.end());
			// debug code: print out sorted saddle points
/*#ifdef I3D_DEBUG
			if (typeid(T) == typeid(i3d::GRAY8))
			{
				for (typename std::vector<std::pair<T, std::pair<size_t, size_t>>>::size_type i = 0; i < sortedSaddlePoints.size(); ++i)
					std::cout << "value " << (int)(sortedSaddlePoints[i].first) << ", basins " << sortedSaddlePoints[i].second.first << ", " << sortedSaddlePoints[i].second.second << std::endl;
			}
			else
			{
				for (typename std::vector<std::pair<T, std::pair<size_t, size_t>>>::size_type i = 0; i < sortedSaddlePoints.size(); ++i)
					std::cout << "value " << sortedSaddlePoints[i].first << ", basins " << sortedSaddlePoints[i].second.first << ", " << sortedSaddlePoints[i].second.second << std::endl;
			}
#endif // #ifdef I3D_DEBUG*/
			// simulate flooding - report when two basins meet
			for (typename std::vector<std::pair<T, std::pair<size_t, size_t> > >::size_type i = 0; i < sortedSaddlePoints.size(); ++i)
				ComputeDynamics(sortedSaddlePoints[i].second.first, sortedSaddlePoints[i].second.second, sortedSaddlePoints[i].first);
		}

		/// returns contour dynamics for the given set of basins touching the contour
		T GetDynamics(const std::set<size_t> &labels)
		{
			T result = 0;
			// TODO try seperate branch for sets with two elements to avoid generality overhead and get some speed
			for (std::set<size_t>::const_iterator first = labels.begin(); first != --labels.end(); ++first)
			{
				std::set<size_t>::const_iterator second = first;
				for (++second; second != labels.end(); ++second)
				{
					// try to lookup the value in already computed ones
					typename std::map<std::pair<size_t, size_t>, T>::iterator singleResult = basinPairsSPDyn.find(std::pair<size_t, size_t>(*first, *second));
					I3DASSERT_BT(singleResult != basinPairsSPDyn.end(), "No already computed dynamics value found (at least the default one) - processing different image than used for building the flooding list?");
					if (result < singleResult->second)
						result = singleResult->second;
					// when no result found, compute it
					else
					{
						SBasin *basin1 = GetBasin(*first);
						SBasin *basin2 = GetBasin(*second);
						// search for the common flooder of the two basins and it's predecessors
						SBasin *flooding1 = basin1;
						SBasin *flooding2 = basin2;
						SBasin *commonFlooder;
						while (true)
						{
							flooding1->visited = flooding2->visited = true;
							if (flooding1->floodedBy)
							{
								if (flooding1->floodedBy->visited)
								{
									commonFlooder = flooding1->floodedBy;
									flooding2 = basin2->GetPredecessorOf(commonFlooder);
									basin1->GetPredecessorOf(commonFlooder);
									break;
								}
								flooding1 = flooding1->floodedBy;
							}
							if (flooding2->floodedBy)
							{
								if (flooding2->floodedBy->visited)
								{
									commonFlooder = flooding2->floodedBy;
									flooding1 = basin1->GetPredecessorOf(commonFlooder);
									basin2->GetPredecessorOf(commonFlooder);
									break;
								}
								flooding2 = flooding2->floodedBy;
							}
							// the case when the flooding list depth is the same from the both sides
							if (flooding1 == flooding2)
							{
								commonFlooder = flooding1;
								flooding1 = basin1->GetPredecessorOf(commonFlooder);
								flooding2 = basin2->GetPredecessorOf(commonFlooder);
								break;
							}
						}
						I3DASSERT_B(flooding1 || flooding2);
						I3DASSERT_B(commonFlooder);
						commonFlooder->ClearVisited();
						// use the higher dynamics found and remember the newly computed value for the later use
						T newResult;
						if (flooding1)
							newResult = flooding1->dynamics;
						else
							newResult = T(0);
						if (flooding2 && flooding2->dynamics > newResult)
							newResult = flooding2->dynamics;
						if (result < newResult)
							result = newResult;
						singleResult->second = newResult;
					}
				}
			}
			return result;
		}
	};

	template <typename T>
	I3D_ALGO_EXPORT void WatershedContourDynamics(const Image3d<T> &input, const Image3d<size_t> &labels, Image3d<T> &saliencyMap, const Neighbourhood &neighbourhood)
	{
		I3DASSERT_ALWAYS_ET(input.GetImageSize() != 0, "Input image should not be empty.");
		I3DASSERT_ALWAYS_ET(input.GetImageSize() == labels.GetImageSize(), "Input and labels image size differs.");

		// scan the input image for the basins and saddle points
		size_t imageSize = input.GetImageSize();
		NeighbourhoodWalker nbhWalker(input.GetSize(), neighbourhood, true);
		std::set<size_t> labelSet;
		CWSCDBasinManager<T> basinManager;
		for (size_t i = 0; i < imageSize; ++i)
		{
			size_t label = labels.GetVoxel(i);
			// for labeled voxels, report values as potential minima
			if (label)
				basinManager.PotentialMinimum(label, input.GetVoxel(i));
			// for watershed lines, report potential saddle points
			else
			{
				labelSet.clear();
				for (nbhWalker.Begin(i); nbhWalker.Valid(); nbhWalker.Next())
				{
					size_t nbhLabel = labels.GetVoxel(nbhWalker.GetIndex());
					if (nbhLabel)
						labelSet.insert(nbhLabel);
				}
				if (labelSet.size() >= 2)
					basinManager.PotentialSaddlePoint(input.GetVoxel(i), labelSet);
				// TODO try to remember the computed label sests (as vector of pointers to a set?) and use them during the contour evaluation - this will cost some memory but may do some speedup
			}
		}

		// construct the flooding list
		basinManager.ComputeFloodingList();

		// evaluate the contours by their dynamics
		saliencyMap.CopyMetaData(input);
		for (size_t i = 0; i < imageSize; ++i)
		{
			size_t label = labels.GetVoxel(i);
			// for labeled voxels, put 0 into the output image
			if (label)
				saliencyMap.SetVoxel(i, T(0));
			// for watershed lines, put the contour dynamics value into the output image
			else
			{
				labelSet.clear();
				for (nbhWalker.Begin(i); nbhWalker.Valid(); nbhWalker.Next())
				{
					size_t nbhLabel = labels.GetVoxel(nbhWalker.GetIndex());
					if (nbhLabel)
						labelSet.insert(nbhLabel);
				}
				if (labelSet.size() >= 2)
					saliencyMap.SetVoxel(i, basinManager.GetDynamics(labelSet));
				else
					saliencyMap.SetVoxel(i, T(0));
			}
		}
	}

	template I3D_ALGO_EXPORT void WatershedContourDynamics(const Image3d<i3d::GRAY8> &input, const Image3d<size_t> &labels, Image3d<i3d::GRAY8> &saliencyMap, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void WatershedContourDynamics(const Image3d<i3d::GRAY16> &input, const Image3d<size_t> &labels, Image3d<i3d::GRAY16> &saliencyMap, const Neighbourhood &neighbourhood);

} // namespace i3d
