/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file watershed_meyer.cc
  \brief Part of mathematical morphology routines - implementation.

  This file contains bodies of mathematical morphology routines dealing with 
  the watershed transformation algorithm by Meyer.
*/ 

#include <vector>
#include <map>
#include <limits>

#include "watersheds.h"
#include "watersheds_internal.h"
#include "i3dassert.h"


#ifdef __GNUG__
#	pragma implementation
#endif

namespace i3d {

	//--------------------------------------------------------------------------
	// Watershed Meyer's algorithm implementation
	//--------------------------------------------------------------------------

	/** Priority queue used in the Meyer's watershed algorithm. Every value can
	be stored only once during the existence of the queue. Whenever a value is
	removed, it's priority defines minimum priority and no inserted element can
	have lower one (in that case, it is	treated as having minimum priority).
	*/
	template <typename T>
	class CWSMeyerQueue
	{
	public:
		/// creates new empty queue - allocates some required space and marks border voxels as already processed
		CWSMeyerQueue(const Image3d<T> &inputExt, size_t paddingSize)
		{
			imageSize = inputExt.GetImageSize();
			inQueue = new bool[imageSize];
			memset(inQueue, 0, inputExt.GetImageSize() * sizeof(bool));
			minimumPriority = (!std::numeric_limits<T>::is_integer && std::numeric_limits<T>::has_denorm == std::denorm_present ? T(0) - std::numeric_limits<T>::max() : std::numeric_limits<T>::min());
			size_t inputExtSize = inputExt.GetImageSize();
			for (size_t i = 0; i < inputExtSize; ++i)
			{
				if (inputExt.OnWideBorder(i, paddingSize))
					inQueue[i] = true;
			}
			// setup the initLevelSize parameter so that memory allocations will reserve initially WSMQ_INIT_LEVEL_SIZE_FLT places for floating point data types, WSMQ_INIT_LEVEL_SIZE_INT_SMALL for small integers data type (data type range smaller than 65537) and WSMQ_INIT_LEVEL_SIZE_INT_LARGE for other integers
#define WSMQ_SMALL_INT_SIZE 65537
#define WSMQ_INIT_LEVEL_SIZE_FLT 4
#define WSMQ_INIT_LEVEL_SIZE_INT_SMALL 16
#define WSMQ_INIT_LEVEL_SIZE_INT_LARGE 8
			if (std::numeric_limits<T>::is_integer)
			{
				if (static_cast<size_t>(std::numeric_limits<T>::max() - std::numeric_limits<T>::min()) < WSMQ_SMALL_INT_SIZE)
					initLevelSize = WSMQ_INIT_LEVEL_SIZE_INT_SMALL;
				else
					initLevelSize = WSMQ_INIT_LEVEL_SIZE_INT_LARGE;
			}
			else
				initLevelSize = WSMQ_INIT_LEVEL_SIZE_FLT;
		}

		/// frees all the allocated resources
		~CWSMeyerQueue()
		{
			delete[] inQueue;
		}

		/// pushes element of value value with priority priority into the queue without checking whether it has been in the queue before
		void PushNoCheck(T priority, const size_t &value)
		{
			I3DASSERT_BT(value < imageSize, "Trying to push value that is higher than allowed.");
			if (minimumPriority > priority)
				priority = minimumPriority;
			SPriorityLevel &priorityLevel = data[priority];
			if (priorityLevel.dataIndexWrite == priorityLevel.dataSize)
			{
				if (!priorityLevel.dataSize)
				{
					// this is a new priority level, allocate data for it
					priorityLevel.dataSize = initLevelSize;
					priorityLevel.data = new size_t[priorityLevel.dataSize];
				}
				else
				{
					// in the case of overflow, check how serious the overflow is and if needed, alocate more space
					if (priorityLevel.dataIndexRead > static_cast<size_t>(priorityLevel.dataSize * 0.75))
					{
						// three quarters of the queue size not needed any more, no space allocation needed
						memcpy(priorityLevel.data, priorityLevel.data + priorityLevel.dataIndexRead, (priorityLevel.dataSize - priorityLevel.dataIndexRead) * sizeof(size_t));
						priorityLevel.dataIndexWrite -= priorityLevel.dataIndexRead;
						priorityLevel.dataIndexRead = 0;
					}
					else
					{
						// queue is still full enough, double it's size
						size_t *newData = new size_t[priorityLevel.dataSize * 2];
						memcpy(newData, priorityLevel.data + priorityLevel.dataIndexRead, (priorityLevel.dataSize - priorityLevel.dataIndexRead) * sizeof(size_t));
						priorityLevel.dataIndexWrite -= priorityLevel.dataIndexRead;
						priorityLevel.dataIndexRead = 0;
						delete[] priorityLevel.data;
						priorityLevel.data = newData;
						priorityLevel.dataSize *= 2;
					}
				}
			}
			priorityLevel.data[priorityLevel.dataIndexWrite++] = value;
			inQueue[value] = true;
		}
		/// pushes element of value value with priority priority into the queue while checking it has not been in the queue before
		void Push(const T &priority, const size_t &value)
		{
			I3DASSERT_BT(value < imageSize, "Trying to push value that is higher than allowed.");
			if (inQueue[value])
				return;
			PushNoCheck(priority, value);
		}
		/// pops the first element with the lowest priority from the queue
		void Pop()
		{
			I3DASSERT_BT(!IsEmpty(), "You are trying to pop an element from an empty queue.");
			I3DASSERT_BT(data.begin()->second.dataIndexRead != data.begin()->second.dataIndexWrite && data.begin()->second.dataIndexRead != data.begin()->second.dataSize, "First priority level of the queue is either empty or uninitialized, queue is corrupted.");
			typename std::map<T, SPriorityLevel>::iterator lowestPriorityPair = data.begin();
			lowestPriorityPair->second.dataIndexRead++;
			if (lowestPriorityPair->second.dataIndexRead == lowestPriorityPair->second.dataIndexWrite)
				data.erase(lowestPriorityPair++);
			if (!data.empty())
				minimumPriority = lowestPriorityPair->first;
		}
		/// returns the first element with the lowest priority from the queue
		size_t Top() const
		{
			I3DASSERT_BT(!IsEmpty(), "You are trying to get an element from an empty queue.");
			I3DASSERT_BT(data.begin()->second.dataIndexRead != data.begin()->second.dataIndexWrite && data.begin()->second.dataIndexRead != data.begin()->second.dataSize, "First priority level of the queue is either empty or uninitialized, queue is corrupted.");
			const SPriorityLevel &priorityLevel = data.begin()->second;
			return priorityLevel.data[priorityLevel.dataIndexRead];
		}
		/// returns whether the queue is empty
		bool IsEmpty() const
		{
			return data.empty();
		}
		/// returns whether has specified value been inside the queue before
		bool HasBeenInside(const size_t &value) const
		{
			I3DASSERT_BT(value < imageSize, "Trying to question value that is higher than allowed.");
			return inQueue[value];
		}

	private:
		/// structure that stores queue for one priority level
		struct SPriorityLevel
		{
			/// constructor that initializes default values
			SPriorityLevel(): dataSize(0), dataIndexRead(0), dataIndexWrite(0), data(0) {}
			/// destructor that frees all the allocated resources
			~SPriorityLevel() { delete[] data; }
			/// maximum number of stored elements
			size_t dataSize;
			/// index of the next reading
			size_t dataIndexRead;
			/// index of the next writing
			size_t dataIndexWrite;
			/// stored values
			size_t *data;
		};
		/// stores the pairs representing the priority and a queue of values
		std::map<T, SPriorityLevel> data;
		/// stores the information whether has specified value been in the queue before
		bool *inQueue;
		/// size of the image we are processing (also size of the inQueue array)
		size_t imageSize;
		/// number of elements created initially for one intensity level queue - this can be used for performace tuning (computed in the constructor) together with realocation parameters in PushNoCheck
		size_t initLevelSize;
		/// minimum priority defining the basket that we are allowed to insert into
		T minimumPriority;
	};

	template <typename T> 
	I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<T> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood)
	{
		I3DASSERT_ALWAYS_ET(input.GetImageSize() != 0, "Input image should not be empty.");
		I3DASSERT_ALWAYS_ET(input.GetImageSize() == marker.GetImageSize(), "Input and marker image size differs.");

		// copy markers into the output
		output.CopyMetaData(input);
		memcpy(output.GetFirstVoxelAddr(), marker.GetFirstVoxelAddr(), marker.GetImageSize() * sizeof(size_t));
		size_t paddingSize = neighbourhood.GetMaxExtent();
		output.PadImage(paddingSize, 0);
		Image3d<T> inputExt(input);
		inputExt.PadImage(paddingSize, std::numeric_limits<T>::max());

		// start growing the output
		// initialize priority queue - insert all the non-labeled voxels (their indices) neighbouring with the so far labeled regions into it
		CWSMeyerQueue<T> voxelsToProcess(inputExt, paddingSize);
		NeighbourhoodWalkerU nbhWalkwer(output.GetSize(), neighbourhood);
		for (size_t i = output.GetImageSize(); i > 0; --i)
		{
			if (output.GetVoxel(i - 1))
			{
				for (nbhWalkwer.Begin(i - 1); nbhWalkwer.Valid(); nbhWalkwer.Next())
				{
					size_t nbhIndex = nbhWalkwer.GetIndex();
					if (output.GetVoxel(nbhIndex) == 0)
						voxelsToProcess.Push(inputExt.GetVoxel(nbhIndex), nbhIndex);
				}
			}
		}
		// while we have some unprocessed voxels in the priority queue, pick one with the lowest voxel value and process it
		std::pair<T, size_t> *voxelsToProcessTmp = new std::pair<T, size_t>[neighbourhood.size()];
		size_t voxelsToProcessTmpElems = 0;
		while (!voxelsToProcess.IsEmpty())
		{
			size_t voxelIndex = voxelsToProcess.Top();
			I3DASSERT_BT(voxelIndex < output.GetImageSize(), "Invalid voxel index in the queue.");
			I3DASSERT_BT(!output.GetVoxel(voxelIndex), "Voxel in the queue should not have label assigned.");
			// check the neighbourhood of the chosen voxel, when it contains just one label, mark it by that label and put all the unmarked neighbours that are (were) not in the queue into it
			bool ambiguousLabel = false;
			for (nbhWalkwer.Begin(voxelIndex); nbhWalkwer.Valid(); nbhWalkwer.Next())
			{
				size_t nbhIndex = nbhWalkwer.GetIndex();
				size_t nbhLabel = output.GetVoxel(nbhIndex);
				if (nbhLabel == 0)
				{
					if (!voxelsToProcess.HasBeenInside(nbhIndex))
					{
						voxelsToProcessTmp[voxelsToProcessTmpElems].first = inputExt.GetVoxel(nbhIndex);
						voxelsToProcessTmp[voxelsToProcessTmpElems++].second = nbhIndex;
					}
					continue;
				}
				else if (output.GetVoxel(voxelIndex) == 0)
				{
					output.SetVoxel(voxelIndex, nbhLabel);
					continue;
				}
				else if (output.GetVoxel(voxelIndex) != nbhLabel)
					ambiguousLabel = true;
			}
			if (ambiguousLabel)
				output.SetVoxel(voxelIndex, 0);
			else
			{
				// in case we were able to chose label for this voxel (there was a single (non-zero) label in the neighbourhood), put all it's non-labeled neighbours into the queue
				if (output.GetVoxel(voxelIndex))
				{
					for (typename std::vector<std::pair<T, size_t> >::size_type i = 0; i != voxelsToProcessTmpElems; ++i)
						voxelsToProcess.PushNoCheck(voxelsToProcessTmp[i].first, voxelsToProcessTmp[i].second);
				}
			}
			voxelsToProcessTmpElems = 0;
			voxelsToProcess.Pop();
		}
		delete[] voxelsToProcessTmp;
		output.RemovePadding(paddingSize);
	}

	template <typename T> 
	I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<T> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood)
	{
		// detect regional minima and construct the marker image
		Image3d<size_t> marker;
		WSInternalMarkerFromMinima(input, marker, neighbourhood);

		// call the watershed
		WatershedMeyer(input, marker, output, neighbourhood);
	}

	template I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<i3d::GRAY8> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<i3d::GRAY16> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<float> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<i3d::GRAY8> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<i3d::GRAY16> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void WatershedMeyer(const Image3d<float> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood);

} // namespace i3d
